import cvm
import dataclasses
import datetime
import ibr
import os
import typing
import zipfile

Indicator = typing.Union[ibr.Indebtedness, ibr.Efficiency, ibr.Profitability, ibr.Valuation]

@dataclasses.dataclass(init=True)
class ReaderResult:
    dfpitr: cvm.DFPITR
    balance_type: cvm.BalanceType
    indicators: typing.List[Indicator]

Reader = typing.Generator[ReaderResult, None, None]

def from_statement(cvm_code: int,
                   reference_date: datetime.date,
                   balance_sheet: cvm.BalanceSheet,
                   income_statement: cvm.IncomeStatement,
                   indicator_types: typing.Iterable[typing.Type[Indicator]]
) -> typing.List[Indicator]:
    indicators = []

    for cls in indicator_types:
        if issubclass(cls, (ibr.Indebtedness, ibr.Profitability)):
            indicators.append(cls.from_statement(balance_sheet, income_statement))
        elif issubclass(cls, ibr.Efficiency):
            indicators.append(cls.from_statement(income_statement))
        elif issubclass(cls, ibr.Valuation):
            indicators.append(cls.from_statement(cvm_code, reference_date, balance_sheet, income_statement))
        else:
            raise TypeError(f'{cls} is not an indicator class')

    return indicators

def reader(file: typing.Union[zipfile.ZipFile, typing.IO, os.PathLike, str],
           indicator_types: typing.Iterable[typing.Type[Indicator]]
) -> Reader:

    dfpitr_reader   = cvm.dfpitr_reader(file, cvm.BalanceFlag.CONSOLIDATED)
    indicator_types = tuple(indicator_types)

    for dfpitr in dfpitr_reader:
        try:
            balance_sheet    = cvm.BalanceSheet.from_dfpitr(dfpitr)
            income_statement = cvm.IncomeStatement.from_dfpitr(dfpitr)
        except ValueError:
            continue
        else:
            indicators = from_statement(
                dfpitr.cvm_code,
                dfpitr.reference_date,
                balance_sheet,
                income_statement,
                indicator_types
            )

            yield ReaderResult(dfpitr, cvm.BalanceType.CONSOLIDATED, indicators)


from __future__ import annotations
import cvm
import dataclasses
import decimal
from ibr.utils import zero_safe_divide, none_safe_divide

@dataclasses.dataclass(init=True)
class Profitability:
    roe: decimal.Decimal
    """Return on Equity."""

    roa: decimal.Decimal
    """Return on Assets."""

    roic: decimal.Decimal
    """Return on Invested Capital."""

    asset_turnover: decimal.Decimal
    """'Giro de Ativo'"""

    @staticmethod
    def from_statement(balance_sheet: cvm.BalanceSheet, income_statement: cvm.IncomeStatement) -> Profitability:
        b = balance_sheet
        i = income_statement

        if b.gross_debt is None:
            equity_plus_gross_debt = None
        else:
            equity_plus_gross_debt = b.equity + b.gross_debt

        return Profitability(
            roe            = zero_safe_divide(i.net_income,            b.equity),
            roa            = zero_safe_divide(i.net_income,            b.total_assets),
            roic           = none_safe_divide(i.ebit - i.tax_expenses, equity_plus_gross_debt),
            asset_turnover = zero_safe_divide(i.revenue,               b.total_assets)
        )


import decimal
import typing

__all__ = [
    'zero_safe_divide',
    'none_safe_divide'
]

def zero_safe_divide(numerator: decimal.Decimal, denominator: decimal.Decimal) -> decimal.Decimal:
    if denominator == 0:
        return 0
    else:
        return numerator / denominator

def none_safe_divide(numerator: typing.Optional[decimal.Decimal],
                     denominator: typing.Optional[decimal.Decimal]
) -> typing.Optional[decimal.Decimal]:

    if numerator is None or denominator is None:
        return None
    else:
        return zero_safe_divide(numerator, denominator)


from __future__ import annotations
import cvm
import dataclasses
import decimal
import typing
from ibr.utils import zero_safe_divide, none_safe_divide

@dataclasses.dataclass(init=True)
class Efficiency:
    gross_margin: decimal.Decimal
    """Margem Bruta"""

    ebitda_margin: typing.Optional[decimal.Decimal]
    """Margem EBITDA"""

    ebit_margin: decimal.Decimal
    """Margem EBIT"""

    net_margin: decimal.Decimal
    """Margem Líquida"""

    @staticmethod
    def from_statement(income_statement: cvm.IncomeStatement) -> Efficiency:
        i = income_statement

        return Efficiency(
            gross_margin  = zero_safe_divide(i.gross_profit, i.revenue),
            ebitda_margin = none_safe_divide(i.ebitda,       i.revenue),
            ebit_margin   = zero_safe_divide(i.ebit,         i.revenue),
            net_margin    = zero_safe_divide(i.net_income,   i.revenue)
        )


from __future__ import annotations
import dataclasses
import decimal
import typing
import cvm
from ibr.utils import zero_safe_divide, none_safe_divide

@dataclasses.dataclass(init=True)
class Indebtedness:
    general_debt: decimal.Decimal
    """'Dívida Geral'"""

    debt_composition: decimal.Decimal
    """'Composição da Dívida' (CD)"""

    net_debt_to_equity: decimal.Decimal
    """'Dívida Líquida/Patrimônio Líquido'"""
    
    net_debt_to_ebitda: typing.Optional[decimal.Decimal]
    """'Dívida Líquida/EBITDA'"""
    
    net_debt_to_ebit: decimal.Decimal
    """'Dívida Líquida/EBIT'"""

    net_equity_to_assets: decimal.Decimal
    """'Dívida Líquida/Ativo'"""
    
    current_ratio: decimal.Decimal
    """'Ativo/Passivo'"""

    @staticmethod
    def from_statement(balance_sheet: cvm.BalanceSheet, income_statement: cvm.IncomeStatement) -> Indebtedness:
        b = balance_sheet
        i = income_statement

        return Indebtedness(
            general_debt         = zero_safe_divide(b.total_liabilities,   b.total_assets),
            debt_composition     = none_safe_divide(b.current_liabilities, b.total_liabilities),
            net_debt_to_equity   = none_safe_divide(b.net_debt,            b.equity),
            net_debt_to_ebitda   = none_safe_divide(b.net_debt,            i.ebitda),
            net_debt_to_ebit     = none_safe_divide(b.net_debt,            i.ebit),
            net_equity_to_assets = zero_safe_divide(b.equity,              b.total_assets),
            current_ratio        = none_safe_divide(b.current_assets,      b.current_liabilities)
        )


from ibr.utils         import *
from ibr.indebtedness  import *
from ibr.efficiency    import *
from ibr.profitability import *
from ibr.valuation     import *
from ibr.reader        import *
from . import _version
__version__ = _version.get_versions()['version']



from setuptools import setup, find_packages
import versioneer

with open('README_en_US.md', mode='r', encoding='utf-8') as f:
    readme_en_us = f.read()

install_requires = [
    'pycvm>=0.3.2',
    'pybov>=0.1.1',
    'yfinance'
]

setup(name                          = 'pyibr',
      version                       = versioneer.get_version(),
      cmdclass                      = versioneer.get_cmdclass(),
      description                   = 'Calculates fundamental analysis indicators of Brazilian companies',
      long_description              = readme_en_us,
      long_description_content_type = 'text/markdown',
      author                        = 'Giovanni L',
      author_email                  = 'callmegiorgio@hotmail.com',
      url                           = 'https://github.com/callmegiorgio/pyibr/',
      license                       = 'MIT',
      packages                      = find_packages(),
      keywords                      = ['investment', 'finances'],
      install_requires              = install_requires,
      python_requires               = '>=3.7'
)


from __future__ import annotations
import b3
import cvm
import dataclasses
import datetime
import decimal
import typing
import yfinance as yf
from ibr.utils import zero_safe_divide, none_safe_divide

@dataclasses.dataclass(init=True)
class Valuation:
    ticker: str
    """'Código de negociação'"""

    quote: decimal.Decimal
    """'Cotação'"""

    outstanding_shares: int
    """'Ações em circulação'"""

    market_capitalization: decimal.Decimal
    """'Valor de Mercado'"""

    enterprise_value: decimal.Decimal
    """'Valor da Firma'"""

    earnings_per_share: decimal.Decimal
    """'Lucro por Ação' (LPA)"""

    book_value_per_share: decimal.Decimal
    """'Valor Patrimonial da Ação' (VPA)"""

    ev_ebitda: typing.Optional[decimal.Decimal]
    """EV/EBITDA"""

    ev_ebit: decimal.Decimal
    """EV/EBIT"""

    @property
    def market_cap(self) -> decimal.Decimal:
        """Same as `market_capitalization`."""

        return self.market_capitalization

    @property
    def ev(self) -> decimal.Decimal:
        """Same as `enterprise_value`."""

        return self.enterprise_value

    @property
    def eps(self) -> decimal.Decimal:
        """Same as `earnings_per_share`."""

        return self.earnings_per_share

    @property
    def bvps(self) -> decimal.Decimal:
        """Same as `book_value_per_share`."""

        return self.book_value_per_share

    @classmethod
    def market_data(cls, cvm_code: int, reference_date: datetime.date) -> typing.Iterable[typing.Tuple[str, decimal.Decimal, int]]:
        raise NotImplementedError(f"method 'market_data' of {cls}")

    @classmethod
    def from_statement(cls,
                       cvm_code: int,
                       reference_date: datetime.date,
                       balance_sheet: cvm.BalanceSheet,
                       income_statement: cvm.IncomeStatement
    ) -> typing.List[Valuation]:
        b = balance_sheet
        i = income_statement

        # Off-market data
        net_debt = b.net_debt or 0
        valuations = []

        # Market data
        for ticker, quote, outstanding_shares in cls.market_data(cvm_code, reference_date):
            market_cap = quote * outstanding_shares
            ev         = market_cap + net_debt

            valuations.append(
                Valuation(
                    ticker                = ticker,
                    quote                 = quote,
                    outstanding_shares    = outstanding_shares,
                    market_capitalization = market_cap,
                    enterprise_value      = ev,
                    earnings_per_share    = zero_safe_divide(i.net_income, outstanding_shares),
                    book_value_per_share  = zero_safe_divide(b.equity,     outstanding_shares),
                    ev_ebitda             = none_safe_divide(ev,           i.ebitda),
                    ev_ebit               = zero_safe_divide(ev,           i.ebit)
                ))

        return valuations

class YfinanceValuation(Valuation):
    @staticmethod
    def market_data_from_ticker(ticker: str, reference_year: int) -> typing.Optional[typing.Tuple[float, int]]:
        start  = f'{reference_year}-01-01'
        end    = f'{reference_year}-12-31'

        quotes = yf.download(ticker, start=start, end=end)

        if quotes.empty:
            return None

        quote  = quotes.iloc[-1]['Close']
        ticker = yf.Ticker(ticker)

        try:
            # Yahoo Finance only stores `sharesOutstanding` for the last
            # quarterly statement provided by the company, so `reference_year`
            # will be ignored...
            outstanding_shares = int(ticker.info['sharesOutstanding'])
        except KeyError:
            return None

        return quote, outstanding_shares

    @classmethod
    def market_data(cls, cvm_code: int, reference_date: datetime.date) -> typing.Iterable[typing.Tuple[str, decimal.Decimal, int]]:
        co = b3.company_detail(cvm_code)

        for codes in co.security_codes:
            data = YfinanceValuation.market_data_from_ticker(codes.ticker + '.SA', reference_date.year)

            if data is None:
                continue

            quote, outstanding_shares = data

            yield (codes.ticker, decimal.Decimal(quote), outstanding_shares)